import sys
import logging
from unittest.mock import (
    patch, Mock
)
from pytest import (
    raises, fixture
)
import pytest

from .test_helper import argv_kiwi_tests

import kiwi

from kiwi.defaults import Defaults
from kiwi.xml_state import XMLState
from kiwi.xml_description import XMLDescription
from kiwi.runtime_checker import RuntimeChecker
from kiwi.exceptions import KiwiRuntimeError


class TestRuntimeChecker:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def setup(self):
        Defaults.set_runtime_checker_metadata(
            '../data/runtime_checker_metadata.yml'
        )
        self.description = XMLDescription(
            '../data/example_runtime_checker_config.xml'
        )
        self.xml_state = XMLState(
            self.description.load()
        )
        self.runtime_config = Mock()
        self.runtime_config.get_oci_archive_tool.return_value = 'umoci'
        kiwi.runtime_checker.RuntimeConfig = Mock(
            return_value=self.runtime_config
        )
        self.runtime_checker = RuntimeChecker(self.xml_state)

    def setup_method(self, cls):
        self.setup()

    @patch('kiwi.runtime_checker.Uri')
    def test_check_image_include_repos_publicly_resolvable(self, mock_Uri):
        uri = Mock()
        uri.is_public.return_value = False
        mock_Uri.return_value = uri
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_image_include_repos_publicly_resolvable()

    @patch('kiwi.runtime_checker.Command.run')
    def test_check_target_dir_on_unsupported_filesystem(self, mock_Command_run):
        stat = Mock()
        stat.output = 'bogus'
        mock_Command_run.return_value = stat
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_target_dir_on_unsupported_filesystem(
                '/some/root_dir'
            )

    def test_invalid_target_dir_pointing_to_shared_cache_1(self):
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_target_directory_not_in_shared_cache(
                '/var/cache//kiwi/foo'
            )

    def test_invalid_target_dir_pointing_to_shared_cache_2(self):
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_target_directory_not_in_shared_cache(
                '/var/cache/kiwi'
            )

    @patch('os.getcwd')
    def test_invalid_target_dir_pointing_to_shared_cache_3(self, mock_getcwd):
        mock_getcwd.return_value = '/'
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_target_directory_not_in_shared_cache(
                'var/cache/kiwi'
            )

    def test_invalid_target_dir_pointing_to_shared_cache_4(self):
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_target_directory_not_in_shared_cache(
                '//var/cache//kiwi/foo'
            )

    @patch('os.path.isdir')
    def test_check_dracut_module_versions_compatible_to_kiwi_no_dracut(
        self, mock_os_path_isdir
    ):
        mock_os_path_isdir.return_value = False
        self.runtime_checker.\
            check_dracut_module_versions_compatible_to_kiwi('root_dir')
        # does not raise if no dracut is installed

    @patch('os.listdir')
    @patch('os.path.isdir')
    @patch('kiwi.runtime_checker.Command.run')
    def test_check_dracut_module_versions_compatible_to_kiwi(
        self, mock_Command_run, mock_os_path_isdir, mock_os_listdir
    ):
        mock_os_path_isdir.return_value = True
        command = Mock()
        command.output = '1.2.3-1.2'
        mock_Command_run.return_value = command
        mock_os_listdir.return_value = ['55kiwi-dump']
        package_manager = Mock()
        package_manager.return_value = 'zypper'
        self.xml_state.get_package_manager = package_manager
        with raises(KiwiRuntimeError) as exception_data:
            self.runtime_checker.\
                check_dracut_module_versions_compatible_to_kiwi('root_dir')
        assert "'dracut-kiwi-oem-dump': 'got:1.2.3, need:>=9.20.1'" in str(
            exception_data.value
        )
        mock_Command_run.assert_called_once_with(
            [
                'chroot', 'root_dir', 'rpm', '-q',
                '--qf', '%{VERSION}', 'dracut-kiwi-oem-dump'
            ]
        )
        package_manager.return_value = 'apt'
        mock_Command_run.reset_mock()
        with raises(KiwiRuntimeError) as exception_data:
            self.runtime_checker.\
                check_dracut_module_versions_compatible_to_kiwi('root_dir')
        assert "'dracut-kiwi-oem-dump': 'got:1.2.3, need:>=9.20.1'" in str(
            exception_data.value
        )
        mock_Command_run.assert_called_once_with(
            [
                'chroot', 'root_dir', 'dpkg-query',
                '-W', '-f', '${Version}', 'dracut-kiwi-oem-dump'
            ]
        )
        mock_Command_run.side_effect = Exception
        # If the package manager call failed with an exception
        # the expected behavior is the runtime check to just pass
        # because there is no data to operate on. Best guess approach
        with self._caplog.at_level(logging.DEBUG):
            self.runtime_checker.\
                check_dracut_module_versions_compatible_to_kiwi('root_dir')

    def test_valid_target_dir_1(self):
        assert self.runtime_checker.check_target_directory_not_in_shared_cache(
            '/var/cache/kiwi-fast-tmpsystemdir'
        ) is None

    def test_valid_target_dir_2(self):
        assert self.runtime_checker.check_target_directory_not_in_shared_cache(
            '/foo/bar'
        ) is None

    def test_check_repositories_configured(self):
        self.xml_state.delete_repository_sections()
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_repositories_configured()

    def test_check_volume_setup_defines_reserved_labels(self):
        xml_state = XMLState(
            self.description.load(), ['vmxSimpleFlavour'], 'oem'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_volume_setup_defines_reserved_labels()

    def test_appx_invalid_name(self):
        xml_state = XMLState(
            self.description.load(), ['wsl_launcher'], 'appx'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_appx_naming_conventions_valid()

    def test_appx_invalid_id(self):
        xml_state = XMLState(
            self.description.load(), ['wsl_id'], 'appx'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_appx_naming_conventions_valid()

    def test_check_volume_setup_defines_multiple_fullsize_volumes(self):
        with raises(KiwiRuntimeError):
            self.runtime_checker.\
                check_volume_setup_defines_multiple_fullsize_volumes()

    def test_check_volume_setup_has_no_root_definition(self):
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_volume_setup_has_no_root_definition()

    def test_check_partuuid_persistency_type_used_with_mbr(self):
        xml_state = XMLState(
            self.description.load(), ['vmxFlavour'], 'iso'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_partuuid_persistency_type_used_with_mbr()

    @patch('kiwi.runtime_checker.CommandCapabilities.has_option_in_help')
    def test_check_luksformat_options_valid(self, mock_has_option_in_help):
        mock_has_option_in_help.return_value = False
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_luksformat_options_valid()

    @patch('kiwi.runtime_checker.Path.which')
    def test_check_container_tool_chain_installed(self, mock_which):
        mock_which.return_value = False
        xml_state = XMLState(
            self.description.load(), ['docker'], 'docker'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_container_tool_chain_installed()

    @patch('kiwi.runtime_checker.Path.which')
    def test_check_container_tool_chain_installed_unknown_tool(
        self, mock_which
    ):
        self.runtime_config.get_oci_archive_tool.return_value = 'budah'
        mock_which.return_value = False
        xml_state = XMLState(
            self.description.load(), ['docker'], 'docker'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_container_tool_chain_installed()

    @patch('kiwi.runtime_checker.Path.which')
    def test_check_container_tool_chain_installed_buildah(
        self, mock_which
    ):
        self.runtime_config.get_oci_archive_tool.return_value = 'buildah'
        mock_which.return_value = False
        xml_state = XMLState(
            self.description.load(), ['docker'], 'docker'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_container_tool_chain_installed()

    @patch('kiwi.runtime_checker.Path.which')
    @patch('kiwi.runtime_checker.CommandCapabilities.check_version')
    def test_check_container_tool_chain_installed_with_version(
        self, mock_cmdver, mock_which
    ):
        mock_which.return_value = True
        mock_cmdver.return_value = False
        xml_state = XMLState(
            self.description.load(), ['docker'], 'docker'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_container_tool_chain_installed()

    @patch('kiwi.runtime_checker.Path.which')
    @patch('kiwi.runtime_checker.CommandCapabilities.check_version')
    @patch('kiwi.runtime_checker.CommandCapabilities.has_option_in_help')
    def test_check_container_tool_chain_installed_with_multitags(
        self, mock_cmdoptions, mock_cmdver, mock_which
    ):
        mock_which.return_value = True
        mock_cmdver.return_value = True
        mock_cmdoptions.return_value = False
        xml_state = XMLState(
            self.description.load(), ['docker'], 'docker'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_container_tool_chain_installed()

    @patch('kiwi.runtime_checker.Defaults.get_boot_image_description_path')
    def test_check_consistent_kernel_in_boot_and_system_image(
        self, mock_boot_path
    ):
        Defaults.set_platform_name('x86_64')
        mock_boot_path.return_value = '../data'
        xml_state = XMLState(
            self.description.load(), ['vmxFlavour'], 'oem'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_consistent_kernel_in_boot_and_system_image()

    def test_check_initrd_selection_required(self):
        description = XMLDescription(
            '../data/example_runtime_checker_no_initrd_system_reference.xml'
        )
        xml_state = XMLState(description.load())
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_initrd_selection_required()

    def test_check_boot_description_exists_no_boot_ref(self):
        description = XMLDescription(
            '../data/example_runtime_checker_no_boot_reference.xml'
        )
        xml_state = XMLState(description.load())
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_boot_description_exists()

    def test_check_boot_description_exists_does_not_exist(self):
        description = XMLDescription(
            '../data/example_runtime_checker_boot_desc_not_found.xml'
        )
        xml_state = XMLState(description.load())
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_boot_description_exists()

    def test_check_xen_uniquely_setup_as_server_or_guest_for_ec2(self):
        self.xml_state.build_type.get_firmware = Mock(
            return_value='ec2'
        )
        self.xml_state.is_xen_server = Mock(
            return_value=True
        )
        self.xml_state.is_xen_guest = Mock(
            return_value=True
        )
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_xen_uniquely_setup_as_server_or_guest()

    def test_check_efi_fat_image_has_correct_size(self):
        self.xml_state.build_type.get_efifatimagesize = Mock(
            return_value='200'
        )
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_efi_fat_image_has_correct_size()

    def test_check_xen_uniquely_setup_as_server_or_guest_for_xen(self):
        self.xml_state.build_type.get_firmware = Mock(
            return_value=None
        )
        self.xml_state.is_xen_server = Mock(
            return_value=True
        )
        self.xml_state.is_xen_guest = Mock(
            return_value=True
        )
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_xen_uniquely_setup_as_server_or_guest()

    def test_check_dracut_module_for_disk_overlay_in_package_list(self):
        xml_state = XMLState(
            self.description.load(), ['vmxFlavour'], 'iso'
        )
        xml_state.build_type.get_overlayroot = Mock(
            return_value=True
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.\
                check_dracut_module_for_disk_overlay_in_package_list()

    @pytest.mark.parametrize("tool_name, tool_binary", [("isomd5sum", "implantisomd5"), ("checkmedia", "tagmedia")])
    @patch('kiwi.runtime_checker.Path.which')
    @patch('kiwi.runtime_checker.RuntimeConfig')
    def test_check_mediacheck_installed_implantisomd5_missing(
        self, mock_RuntimeConfig, mock_which, tool_name, tool_binary
    ):
        runtime_config = Mock()
        runtime_config.get_iso_media_tag_tool.return_value = tool_name
        mock_RuntimeConfig.return_value = runtime_config

        mock_which.return_value = False
        xml_state = XMLState(
            self.description.load(), ['vmxFlavour'], 'iso'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError) as rt_err_ctx:
            runtime_checker.check_mediacheck_installed()

        assert f"Required tool {tool_binary} not found in caller environment" in str(rt_err_ctx.value)

    def test_check_dracut_module_for_live_iso_in_package_list(self):
        xml_state = XMLState(
            self.description.load(), ['vmxFlavour'], 'iso'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_dracut_module_for_live_iso_in_package_list()

    def test_check_dracut_module_for_disk_oem_in_package_list(self):
        xml_state = XMLState(
            self.description.load(), ['vmxFlavour'], 'oem'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_dracut_module_for_disk_oem_in_package_list()

    def test_check_dracut_module_for_oem_install_in_package_list(self):
        xml_state = XMLState(
            self.description.load(), ['vmxFlavour'], 'oem'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.\
                check_dracut_module_for_oem_install_in_package_list()

    def test_check_volume_label_used_with_lvm(self):
        with raises(KiwiRuntimeError):
            self.runtime_checker.check_volume_label_used_with_lvm()

    def test_check_swap_name_used_with_lvm(self):
        xml_state = XMLState(
            self.description.load(), ['vmxFlavour'], 'oem'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_swap_name_used_with_lvm()

    def test_check_preferences_data_no_version(self):
        xml_state = XMLState(
            self.description.load(), ['docker'], 'docker'
        )
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_image_version_provided()

    def test_check_image_type_unique(self):
        description = XMLDescription(
            '../data/example_runtime_checker_conflicting_types.xml'
        )
        xml_state = XMLState(description.load())
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_image_type_unique()

    def test_check_include_references_unresolvable(self):
        description = XMLDescription(
            '../data/example_runtime_checker_include_nested_reference.xml'
        )
        xml_state = XMLState(description.load())
        runtime_checker = RuntimeChecker(xml_state)
        with raises(KiwiRuntimeError):
            runtime_checker.check_include_references_unresolvable()

    def test_check_package_in_list(self):
        assert RuntimeChecker._package_in_list(['a', 'b'], ['b']) == 'b'
        assert RuntimeChecker._package_in_list(['a', 'b'], ['c']) == ''

    def teardown(self):
        sys.argv = argv_kiwi_tests

    def teardown_method(self, cls):
        self.teardown()



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
# project
import importlib
from typing import List
from abc import (
    ABCMeta,
    abstractmethod
)

from kiwi.xml_state import XMLState

from kiwi.exceptions import KiwiBootImageSetupError


class BootImage(metaclass=ABCMeta):
    """
    **BootImge Factory**

    :param object xml_state: Instance of :class:`XMLState`
    :param string target_dir: target dir to store the initrd
    :param string root_dir: system image root directory
    :param list signing_keys: list of package signing keys
    """
    @abstractmethod
    def __init__(self) -> None:
        return None  # pragma: no cover

    @staticmethod
    def new(
        xml_state: XMLState, target_dir: str,
        root_dir: str = None, signing_keys: List = None
    ):
        initrd_system = xml_state.get_initrd_system()
        name_map = {
            'kiwi': {'builtin_kiwi': 'BootImageKiwi'},
            'dracut': {'dracut': 'BootImageDracut'},
            'none': {'base': 'BootImageBase'}
        }
        try:
            (boot_image_namespace, boot_image_name) = \
                list(name_map[initrd_system].items())[0]
            boot_image = importlib.import_module(
                'kiwi.boot.image.{0}'.format(boot_image_namespace)
            )
            return boot_image.__dict__[boot_image_name](
                xml_state, target_dir, root_dir, signing_keys
            )
        except Exception as issue:
            raise KiwiBootImageSetupError(
                'Support for {0} initrd system not implemented: {1}'.format(
                    initrd_system, issue
                )
            )



from pytest import raises
from unittest.mock import patch

from kiwi.container.setup.wsl import ContainerSetupWsl
from kiwi.exceptions import KiwiContainerSetupError


class TestContainerSetupAppx:
    @patch('os.path.exists')
    def setup(self, mock_exists):
        self.wsl = ContainerSetupWsl('root_dir')

    @patch('os.path.exists')
    def setup_method(self, cls, mock_exists):
        self.setup()

    @patch('os.path.exists')
    def test_setup(self, mock_exists):

        def exists(path):
            if path == path_to_fail:
                return False
            return True

        mock_exists.side_effect = exists

        path_to_fail = 'root_dir/etc/wsl-distribution.conf'
        with raises(KiwiContainerSetupError):
            self.wsl.setup()

        path_to_fail = 'root_dir/etc/wsl.conf'
        with raises(KiwiContainerSetupError):
            self.wsl.setup()



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import re
import logging
from lxml import etree
from urllib.parse import (
    urlparse, ParseResult, quote
)
from urllib.request import urlopen
from urllib.request import Request
import requests
from uuid import uuid4
from typing import Optional

# project
from kiwi.defaults import Defaults
from kiwi.runtime_config import RuntimeConfig

from kiwi.exceptions import (
    KiwiUriStyleUnknown,
    KiwiUriTypeUnknown,
    KiwiUriOpenError
)

log = logging.getLogger('kiwi')


class Uri:
    """
    **Normalize and manage URI types**
    """
    def __init__(
        self, uri: str, repo_type: str = 'rpm-md', source_type: str = ''
    ):
        """
        Manage kiwi source URIs and allow transformation into
        standard URLs

        :param str uri: URI, remote, local or metalink repository location
            The resource type as part of the URI can be set to one of:
            * http:
            * https:
            * ftp:
            * obs:
            * dir:
            * file:
            * obsrepositories:
            * this:
            The special this:// type resolve to the image description
            directory. The code to resolve this is not part of the Uri
            class because it has no state information about the image
            description directory. Therefore the resolving of the this://
            path happens on construction of an XMLState object as
            part of the resolve_this_path() method. The method resolves
            the path into a native dir:// URI which can be properly
            handled here.
        :param str repo_type:
            repository type name, defaults to 'rpm-md' and is only
            effectively used when building inside of the open
            build service which maps local repositories to a
            specific environment
        :param str source_type:
            specify source type if the provided URI is a service.
            Currently only the metalink source type is handled
        """
        self.runtime_config = RuntimeConfig()

        if source_type == 'metalink':
            uri = self._resolve_metalink_uri(uri)

        self.repo_type = repo_type
        self.uri = uri if not uri.startswith(os.sep) else ''.join(
            [Defaults.get_default_uri_type(), uri]
        )

        self.remote_uri_types = {
            'http': True,
            'https': True,
            'ftp': True,
            'obs': True
        }
        self.local_uri_type = {
            'dir': True,
            'file': True,
            'obsrepositories': True
        }

    def translate(self, check_build_environment: bool = True) -> str:
        """
        Translate repository location according to their URI type

        Depending on the URI type the provided location needs to
        be adapted e.g updated by the service URL in case of an
        open buildservice project name

        :raises KiwiUriStyleUnknown: if the uri scheme can't be detected, is
            unknown or it is inconsistent with the build environment
        :param bool check_build_environment: specify if the uri translation
            should depend on the environment the build is called in. As of
            today this only effects the translation result if the image
            build happens inside of the Open Build Service

        :return: translated repository location

        :rtype: str
        """
        uri = urlparse(self.uri)
        if not uri.scheme:
            raise KiwiUriStyleUnknown(
                'URI scheme not detected {uri}'.format(uri=self.uri)
            )
        elif uri.scheme == 'obs':
            if check_build_environment and Defaults.is_buildservice_worker():
                return self._buildservice_path(
                    name=''.join([uri.netloc, uri.path]).replace(':/', ':'),
                    fragment=uri.fragment,
                    urischeme=uri.scheme
                )
            else:
                return self._obs_project_download_link(
                    ''.join([uri.netloc, uri.path]).replace(':/', ':')
                )
        elif uri.scheme == 'obsrepositories':
            if not Defaults.is_buildservice_worker():
                raise KiwiUriStyleUnknown(
                    'Only the buildservice can use the {0} schema'.format(
                        uri.scheme
                    )
                )
            return self._buildservice_path(
                name=''.join([uri.netloc, uri.path]).replace(':/', ':'),
                fragment=uri.fragment,
                urischeme=uri.scheme
            )
        elif uri.scheme == 'dir':
            return self._local_path(uri.path)
        elif uri.scheme == 'file':
            return self._local_path(uri.path)
        elif uri.scheme.startswith('http') or uri.scheme == 'ftp':
            netloc = uri.netloc
            uri_with_credentials_pattern = '^(.*):(.*)@(.*)$'
            sensitive_match = re.match(uri_with_credentials_pattern, netloc)
            if sensitive_match:
                netloc = "{0}:{1}@{2}".format(
                    quote(sensitive_match.group(1)),
                    quote(sensitive_match.group(2)),
                    sensitive_match.group(3)
                )
            if self._get_credentials_uri() or not uri.query:
                return ''.join(
                    [uri.scheme, '://', netloc, uri.path]
                )
            else:
                return ''.join(
                    [uri.scheme, '://', netloc, uri.path, '?', uri.query]
                )
        else:
            raise KiwiUriStyleUnknown(
                f'URI schema {self.uri} not supported'
            )

    @staticmethod
    def print_sensitive(location: str) -> str:
        uri_with_credentials_pattern = '^.*://(.*:.*)@.*'
        sensitive_match = re.match(uri_with_credentials_pattern, location)
        if sensitive_match:
            return location.replace(sensitive_match.group(1), '******')
        else:
            return location

    def credentials_file_name(self) -> str:
        """
        Filename to store repository credentials

        :return: credentials file name

        :rtype: str
        """
        uri = self._get_credentials_uri()
        # initialize query with default credentials file name.
        # The information will be overwritten if the uri contains
        # a parameter query with a credentials parameter
        query = {'credentials': 'kiwiRepoCredentials'}

        if uri:
            query = dict(params.split('=') for params in uri.query.split('&'))  # type: ignore

        return query['credentials']

    def alias(self) -> str:
        """
        Create hex representation of uuid4

        If the repository definition from the XML description does
        not provide an alias, kiwi creates one for you. However it's
        better to assign a human readable alias in the XML
        configuration

        :return: alias name as hex representation of uuid4

        :rtype: str
        """
        return uuid4().hex

    def is_remote(self) -> bool:
        """
        Check if URI is a remote or local location

        :return: True|False

        :rtype: bool
        """
        uri = urlparse(self.uri)
        if not uri.scheme:
            raise KiwiUriStyleUnknown(
                f'URI scheme not detected {self.uri}'
            )
        if uri.scheme == 'obs' and Defaults.is_buildservice_worker():
            return False
        elif uri.scheme in self.remote_uri_types:
            return True
        elif uri.scheme in self.local_uri_type:
            return False
        else:
            raise KiwiUriTypeUnknown(
                f'URI type {uri.scheme} unknown'
            )

    def is_public(self) -> bool:
        """
        Check if URI is considered to be publicly reachable

        :return: True|False

        :rtype: bool
        """
        uri = urlparse(self.uri)
        if not uri.scheme:
            # unknown uri schema is considered not public
            return False
        elif uri.scheme == 'obs':
            # obs is public but only if the configured download_server is public
            return self.runtime_config.is_obs_public()
        elif uri.scheme in self.remote_uri_types:
            # listed in remote uri types, thus public
            return True
        else:
            # unknown uri type considered not public
            return False

    def get_fragment(self) -> str:
        """
        Returns the fragment part of the URI.

        :return: fragment part of the URI if any, empty string otherwise

        :rtype: str
        """
        uri = urlparse(self.uri)
        return uri.fragment

    def _get_credentials_uri(self) -> Optional[ParseResult]:
        uri = urlparse(self.uri)
        credentials_uri = None
        if uri.query and uri.query.startswith('credentials='):
            credentials_uri = uri
        return credentials_uri

    def _local_path(self, path: str) -> str:
        return os.path.abspath(os.path.normpath(path))

    def _obs_project_download_link(self, name: str) -> str:
        name_parts = name.split(os.sep)
        repository = name_parts.pop()
        project = os.sep.join(name_parts)
        download_link = None
        try:
            download_link = os.sep.join(
                [
                    self.runtime_config.get_obs_download_server_url(),
                    project.replace(':', ':/'), repository
                ]
            )
            if not Defaults.is_buildservice_worker():
                request = requests.get(download_link)
                request.raise_for_status()
                return request.url
            else:
                log.warning(
                    'Using {0} without location verification due to build '
                    'in isolated environment'.format(download_link)
                )
                return download_link
        except Exception as issue:
            raise KiwiUriOpenError(
                f'{download_link}: {issue}'
            )

    def _buildservice_path(
        self, name: str, urischeme: str, fragment: str = ''
    ) -> str:
        """
        Special to openSUSE buildservice. If the buildservice builds
        the image it arranges the repos for each build in a special
        environment, the so called build worker.
        """
        bs_source_dir = '/usr/src/packages/SOURCES'
        if self.repo_type == 'container':
            if urischeme == 'obsrepositories':
                local_path = os.sep.join(
                    [bs_source_dir, 'containers/_obsrepositories', name]
                )
            else:
                local_path = os.sep.join(
                    [bs_source_dir, 'containers', name]
                )
            if fragment:
                local_path = ''.join([local_path, '#', fragment])
        else:
            local_path = os.sep.join(
                [bs_source_dir, 'repos', name]
            )
        return self._local_path(local_path)

    def _resolve_metalink_uri(self, uri: str) -> str:
        selected_repo_source = uri
        namespace_map = dict(
            metalink="http://www.metalinker.org/"
        )
        expression = '//metalink:file[@name="repomd.xml"]/metalink:resources/*'
        try:
            metalink_location = urlopen(Request(uri))
            xml = etree.parse(metalink_location)
            url_list = xml.getroot().xpath(
                expression, namespaces=namespace_map
            )
            source_dict = {}
            for url in url_list:
                if url.get('protocol') == 'https':
                    source_dict[url.text] = int(url.get('preference'))
            start_preference = 0
            for url in sorted(source_dict.keys()):
                preference = source_dict[url]
                if preference > start_preference:
                    selected_repo_source = url
                    start_preference = preference
        except Exception as issue:
            raise KiwiUriOpenError(
                f'Failed to resolve metalink URI {uri}: {issue}'
            )
        selected_repo_source = selected_repo_source.replace(
            'repodata/repomd.xml', ''
        )
        return selected_repo_source



from pytest_container import auto_container_parametrize, DerivedContainer


FUNCTIONS_COPY_CONTAINERFILE = """COPY kiwi/config/functions.sh /bin/functions.sh
"""


CONTAINERS = [
    DerivedContainer(base=url, containerfile=FUNCTIONS_COPY_CONTAINERFILE)
    for url in [
        "registry.opensuse.org/opensuse/leap:15.5",
        "registry.opensuse.org/opensuse/tumbleweed:latest",
        "registry.fedoraproject.org/fedora:latest"
    ]
]

(
    LEAP_15_5,
    TUMBLEWEED,
    FEDORA
) = CONTAINERS


CONTAINERS_WITH_ZYPPER = [
    LEAP_15_5,
    TUMBLEWEED
]


CONTAINERS_WITH_DNF = [
    FEDORA
]


def pytest_generate_tests(metafunc):
    auto_container_parametrize(metafunc)



from unittest.mock import (
    patch, call, MagicMock
)
import io
import unittest.mock as mock

from kiwi.repository.apt import RepositoryApt


class TestRepositoryApt:
    @patch('kiwi.repository.apt.Temporary.unmanaged_file')
    @patch('kiwi.repository.apt.PackageManagerTemplateAptGet')
    @patch('kiwi.repository.apt.Path.create')
    def setup(self, mock_path, mock_template, mock_temp):
        self.apt_conf = mock.Mock()
        mock_template.return_value = self.apt_conf

        template = mock.Mock()
        self.apt_conf.get_host_template.return_value = template

        tmpfile = mock.Mock()
        tmpfile.name = 'tmpfile'
        mock_temp.return_value = tmpfile
        root_bind = mock.Mock()
        root_bind.root_dir = '../data'
        root_bind.shared_location = '/shared-dir'

        with patch('builtins.open', create=True):
            self.repo = RepositoryApt(
                root_bind, custom_args=['exclude_docs']
            )

            self.exclude_docs = True
            self.apt_conf.get_host_template.assert_called_once_with(
                self.exclude_docs
            )
            template.substitute.assert_called_once_with(
                {
                    'apt_shared_base': '/shared-dir/apt-get',
                    'unauthenticated': 'true'
                }
            )
            repo = RepositoryApt(
                root_bind, custom_args=['check_signatures']
            )
            assert repo.custom_args == []
            assert repo.unauthenticated == 'false'

            repo = RepositoryApt(root_bind)
            assert repo.custom_args == []
            assert repo.unauthenticated == 'true'

    @patch('kiwi.repository.apt.Temporary.unmanaged_file')
    @patch('kiwi.repository.apt.PackageManagerTemplateAptGet')
    @patch('kiwi.repository.apt.Path.create')
    def setup_method(self, cls, mock_path, mock_template, mock_temp):
        self.setup()

    def test_use_default_location(self):
        template = mock.Mock()
        template.substitute.return_value = 'template-data'
        self.apt_conf.get_image_template.return_value = template
        with patch('builtins.open', create=True):
            self.repo.use_default_location()
        assert self.repo.shared_apt_get_dir['sources-dir'] == \
            '../data/etc/apt/sources.list.d'
        assert self.repo.shared_apt_get_dir['preferences-dir'] == \
            '../data/etc/apt/preferences.d'
        self.apt_conf.get_image_template.assert_called_once_with(
            self.exclude_docs
        )
        template.substitute.assert_called_once_with(
            {'apt_shared_base': '../data/etc/apt', 'unauthenticated': 'true'}
        )

    def test_runtime_config(self):
        assert self.repo.runtime_config()['apt_get_args'] == \
            self.repo.apt_get_args
        assert self.repo.runtime_config()['command_env'] == \
            self.repo.command_env

    def test_setup_package_database_configuration(self):
        # just pass
        self.repo.setup_package_database_configuration()

    @patch('os.path.exists')
    @patch('kiwi.command.Command.run')
    def test_add_repo_with_priority(self, mock_Command_run, mock_exists):
        mock_exists.return_value = True
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.repo.add_repo(
                'foo', '/srv/my-repo', 'deb', '42', 'xenial', 'a b',
                customization_script='custom_script'
            )
            assert mock_open.call_args_list == [
                call('/shared-dir/apt-get/sources.list.d/foo.sources', 'w'),
                call('/shared-dir/apt-get/preferences.d/foo.pref', 'w')
            ]
            assert file_handle.write.call_args_list == [
                call(
                    'Types: deb\n'
                    'URIs: file:/srv/my-repo\n'
                    'Suites: xenial\n'
                    'Components: a b\n'
                ),
                call('Package: *\n'),
                call('Pin: origin ""\n'),
                call('Pin-Priority: 42\n')
            ]
            assert mock_Command_run.call_args_list == [
                call(
                    [
                        'bash', '--norc', 'custom_script',
                        '/shared-dir/apt-get/sources.list.d/foo.sources'
                    ]
                ),
                call(
                    [
                        'bash', '--norc', 'custom_script',
                        '/shared-dir/apt-get/preferences.d/foo.pref'
                    ]
                )
            ]
        mock_exists.return_value = False
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.repo.add_repo(
                'foo',
                'http://download.opensuse.org/repositories/V:/A:/C/Debian_9.0/',
                'deb', '99', 'xenial', 'a b'
            )
            assert file_handle.write.call_args_list == [
                call(
                    'Types: deb\n'
                    'URIs: http://download.opensuse.org/repositories/V:/A:/C/Debian_9.0/\n'
                    'Suites: xenial\n'
                    'Components: a b\n'
                ),
                call('Package: *\n'),
                call('Pin: origin "download.opensuse.org"\n'),
                call('Pin-Priority: 99\n')
            ]

    @patch('os.path.exists')
    def test_add_repo_distribution(self, mock_exists):
        mock_exists.return_value = True
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.repo.add_repo(
                'foo', 'kiwi_iso_mount/uri', 'deb', None, 'xenial', 'a b',
                architectures='amd64,arm64'
            )
            file_handle.write.assert_called_once_with(
                'Types: deb\n'
                'URIs: file:/kiwi_iso_mount/uri\n'
                'Architectures: amd64 arm64\n'
                'Suites: xenial\n'
                'Components: a b\n'
            )
            mock_open.assert_called_once_with(
                '/shared-dir/apt-get/sources.list.d/foo.sources', 'w'
            )

    @patch('os.path.exists')
    def test_add_repo_distribution_without_gpgchecks(self, mock_exists):
        mock_exists.return_value = True
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.repo.add_repo(
                'foo', 'kiwi_iso_mount/uri', 'deb', None, 'xenial', 'a b',
                repo_gpgcheck=False, pkg_gpgcheck=False
            )
            file_handle.write.assert_called_once_with(
                'Types: deb\n'
                'URIs: file:/kiwi_iso_mount/uri\n'
                'Suites: xenial\n'
                'Components: a b\n'
                'trusted: yes\n'
                'check-valid-until: no\n'
            )
            mock_open.assert_called_once_with(
                '/shared-dir/apt-get/sources.list.d/foo.sources', 'w'
            )

    @patch('os.path.exists')
    def test_add_repo_distribution_default_component(self, mock_exists):
        mock_exists.return_value = True
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.repo.add_repo(
                'foo', '/kiwi_iso_mount/uri', 'deb', None, 'xenial'
            )
            file_handle.write.assert_called_once_with(
                'Types: deb\n'
                'URIs: file:/kiwi_iso_mount/uri\n'
                'Suites: xenial\n'
                'Components: main\n'
            )
            mock_open.assert_called_once_with(
                '/shared-dir/apt-get/sources.list.d/foo.sources', 'w'
            )

    @patch('os.path.exists')
    def test_add_repo_flat(self, mock_exists):
        mock_exists.return_value = False
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.repo.add_repo(
                'foo', 'http://repo.com', 'deb'
            )
            file_handle.write.assert_called_once_with(
                'Types: deb\n'
                'URIs: http://repo.com\n'
                'Suites: ./\n'
            )
            mock_open.assert_called_once_with(
                '/shared-dir/apt-get/sources.list.d/foo.sources', 'w'
            )

    @patch('kiwi.repository.apt.os.unlink')
    @patch('kiwi.repository.apt.os.path.exists')
    @patch('kiwi.repository.apt.Command.run')
    def test_import_trusted_keys(self, mock_run, mock_exists, mock_unlink):
        mock_exists.return_value = True
        self.repo.import_trusted_keys(['key-file-a.asc', 'key-file-b.asc'])
        assert mock_run.call_args_list == [
            call([
                'gpg', '--no-options', '--no-default-keyring',
                '--no-auto-check-trustdb', '--trust-model', 'always',
                '--keyring', '/shared-dir/apt-get/trusted-keybox.gpg',
                '--import', '--ignore-time-conflict', 'key-file-a.asc'
            ]), call([
                'gpg', '--no-options', '--no-default-keyring',
                '--no-auto-check-trustdb', '--trust-model', 'always',
                '--keyring', '/shared-dir/apt-get/trusted-keybox.gpg',
                '--import', '--ignore-time-conflict', 'key-file-b.asc'
            ]), call([
                'gpg', '--no-options', '--no-default-keyring',
                '--no-auto-check-trustdb', '--trust-model', 'always',
                '--keyring', '/shared-dir/apt-get/trusted-keybox.gpg',
                '--export', '--yes', '--output',
                '/shared-dir/apt-get/trusted.gpg'
            ])
        ]

    @patch('kiwi.path.Path.wipe')
    def test_delete_repo(self, mock_wipe):
        self.repo.delete_repo('foo')
        mock_wipe.assert_called_once_with(
            '/shared-dir/apt-get/sources.list.d/foo.sources'
        )

    @patch('kiwi.path.Path.wipe')
    @patch('os.walk')
    def test_cleanup_unused_repos(self, mock_walk, mock_path):
        mock_walk.return_value = [
            ('/foo', ('bar', 'baz'), ('spam', 'eggs'))
        ]
        self.repo.repo_names = ['eggs']
        self.repo.cleanup_unused_repos()
        mock_path.assert_called_once_with(
            '/shared-dir/apt-get/sources.list.d/spam'
        )

    @patch('kiwi.path.Path.wipe')
    @patch('kiwi.path.Path.create')
    def test_delete_all_repos(self, mock_create, mock_wipe):
        self.repo.delete_all_repos()
        mock_wipe.assert_called_once_with(
            '/shared-dir/apt-get/sources.list.d'
        )
        mock_create.assert_called_once_with(
            '/shared-dir/apt-get/sources.list.d'
        )

    @patch('kiwi.path.Path.wipe')
    def test_delete_repo_cache(self, mock_wipe):
        self.repo.delete_repo_cache('foo')
        assert mock_wipe.call_args_list == [
            call('/shared-dir/apt-get/archives'),
            call('/shared-dir/apt-get/pkgcache.bin'),
            call('/shared-dir/apt-get/srcpkgcache.bin')
        ]

    @patch('os.path.isfile')
    @patch('os.unlink')
    def test_cleanup(self, mock_os_unlink, mock_os_path_isfile):
        self.repo.cleanup()
        mock_os_unlink.assert_called_once_with('tmpfile')



import logging
import sys
import os
from pytest import fixture
from unittest.mock import (
    patch, call, Mock, MagicMock
)

import kiwi

from ..test_helper import argv_kiwi_tests

from kiwi.tasks.system_build import SystemBuildTask


class TestSystemBuildTask:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def setup(self):
        sys.argv = [
            sys.argv[0], '--profile', 'vmxFlavour', 'system', 'build',
            '--description', '../data/description',
            '--target-dir', 'some-target'
        ]
        self.abs_target_dir = os.path.abspath('some-target')

        kiwi.tasks.system_build.Privileges = Mock()
        kiwi.tasks.system_build.Path = Mock()

        kiwi.tasks.system_build.Help = Mock(
            return_value=Mock()
        )

        self.runtime_checker = Mock()
        kiwi.tasks.base.RuntimeChecker = Mock(
            return_value=self.runtime_checker
        )

        self.setup = Mock()
        kiwi.tasks.system_build.SystemSetup = Mock(
            return_value=self.setup
        )

        self.profile = Mock()
        self.profile.dot_profile = dict()
        kiwi.tasks.system_build.Profile = Mock(
            return_value=self.profile
        )

        self.result = Mock()
        self.builder = MagicMock()
        self.builder.create = Mock(
            return_value=self.result
        )
        kiwi.tasks.system_build.ImageBuilder.new = Mock(
            return_value=self.builder
        )

        self.task = SystemBuildTask()
        self.task.runtime_config = MagicMock()

    def setup_method(self, cls):
        self.setup()

    def teardown(self):
        sys.argv = argv_kiwi_tests

    def teardown_method(self, cls):
        self.teardown()

    def _init_command_args(self):
        self.task.command_args = {}
        self.task.command_args['help'] = False
        self.task.command_args['build'] = False
        self.task.command_args['--set-type-attr'] = ['volid=some']
        self.task.command_args['--set-release-version'] = '99'
        self.task.command_args['--allow-existing-root'] = True
        self.task.command_args['--description'] = '../data/description'
        self.task.command_args['--target-dir'] = 'some-target'
        self.task.command_args['--set-repo'] = None
        self.task.command_args['--set-repo-credentials'] = None
        self.task.command_args['--add-repo'] = []
        self.task.command_args['--add-repo-credentials'] = []
        self.task.command_args['--add-package'] = []
        self.task.command_args['--add-bootstrap-package'] = []
        self.task.command_args['--delete-package'] = []
        self.task.command_args['--ignore-repos'] = False
        self.task.command_args['--ignore-repos-used-for-build'] = False
        self.task.command_args['--set-container-derived-from'] = None
        self.task.command_args['--set-container-tag'] = None
        self.task.command_args['--add-container-label'] = []
        self.task.command_args['--clear-cache'] = False
        self.task.command_args['--signing-key'] = []

    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.xml_state.XMLState.get_repositories_signing_keys')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build(
        self, mock_SystemPrepare, mock_keys, mock_log
    ):
        manager = MagicMock()
        system_prepare = Mock()
        system_prepare.setup_repositories = Mock(
            return_value=manager
        )
        mock_SystemPrepare.return_value.__enter__.return_value = system_prepare
        mock_keys.return_value = ['some_key', 'some_other_key']
        self._init_command_args()
        self.task.command_args['build'] = True
        self.task.process()
        self.runtime_checker.\
            check_boot_description_exists.assert_called_once_with()
        self.runtime_checker.\
            check_initrd_selection_required.assert_called_once_with()
        self.runtime_checker.\
            check_consistent_kernel_in_boot_and_system_image.\
            assert_called_once_with()
        self.runtime_checker.\
            check_container_tool_chain_installed.assert_called_once_with()
        self.runtime_checker.\
            check_volume_setup_defines_multiple_fullsize_volumes.\
            assert_called_once_with()
        self.runtime_checker.\
            check_volume_setup_has_no_root_definition.\
            assert_called_once_with()
        self.runtime_checker.\
            check_volume_label_used_with_lvm.assert_called_once_with()
        self.runtime_checker.\
            check_swap_name_used_with_lvm.assert_called_once_with()
        self.runtime_checker.\
            check_xen_uniquely_setup_as_server_or_guest.\
            assert_called_once_with()
        self.runtime_checker.\
            check_target_directory_not_in_shared_cache.\
            assert_called_once_with(self.abs_target_dir)
        self.runtime_checker.\
            check_dracut_module_versions_compatible_to_kiwi.\
            assert_called_once_with(
                self.abs_target_dir + '/build/image-root'
            )
        self.runtime_checker.\
            check_mediacheck_installed.assert_called_once_with()
        self.runtime_checker.\
            check_dracut_module_for_live_iso_in_package_list.\
            assert_called_once_with()
        self.runtime_checker.\
            check_repositories_configured.assert_called_once_with()
        self.runtime_checker.\
            check_dracut_module_for_disk_overlay_in_package_list.\
            assert_called_once_with()
        self.runtime_checker.\
            check_dracut_module_for_disk_oem_in_package_list.\
            assert_called_once_with()
        self.runtime_checker.\
            check_dracut_module_for_oem_install_in_package_list.\
            assert_called_once_with()
        system_prepare.setup_repositories.assert_called_once_with(
            False, ['some_key', 'some_other_key'], None
        )
        system_prepare.install_bootstrap.assert_called_once_with(
            manager.__enter__.return_value, []
        )
        system_prepare.install_system.assert_called_once_with(
            manager.__enter__.return_value
        )
        self.profile.create.assert_called_once_with(
            self.abs_target_dir + '/build/image-root/.profile'
        )
        self.setup.import_description.assert_called_once_with()
        self.setup.import_overlay_files.assert_called_once_with()
        self.setup.import_repositories_marked_as_imageinclude.\
            assert_called_once_with()
        self.setup.call_post_bootstrap_script.assert_called_once_with()
        self.setup.call_config_script.assert_called_once_with()
        self.setup.import_image_identifier.assert_called_once_with()
        self.setup.setup_groups.assert_called_once_with()
        self.setup.setup_users.assert_called_once_with()
        self.setup.setup_keyboard_map.assert_called_once_with()
        self.setup.setup_locale.assert_called_once_with()
        self.setup.setup_plymouth_splash.assert_called_once_with()
        self.setup.setup_timezone.assert_called_once_with()
        self.setup.setup_permissions.assert_called_once_with()
        self.setup.import_files.assert_called_once_with()
        self.setup.setup_registry_import.assert_called_once_with()
        self.setup.setup_selinux_file_contexts.assert_called_once_with()
        system_prepare.pinch_system.assert_has_calls(
            [call(force=False), call(force=True)]
        )
        self.setup.create_system_files.assert_called_once_with()
        assert system_prepare.clean_package_manager_leftovers.called
        self.setup.call_image_script.assert_called_once_with()
        self.builder.create.assert_called_once_with()
        self.result.print_results.assert_called_once_with()
        self.result.dump.assert_called_once_with(
            os.sep.join([self.abs_target_dir, 'kiwi.result'])
        )

    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.xml_state.XMLState.get_repositories_signing_keys')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_add_package(
        self, mock_SystemPrepare, mock_keys, mock_log
    ):
        manager = MagicMock()
        system_prepare = Mock()
        system_prepare.setup_repositories = Mock(
            return_value=manager
        )
        mock_SystemPrepare.return_value.__enter__.return_value = system_prepare
        mock_keys.return_value = ['some_key', 'some_other_key']
        self._init_command_args()
        self.task.command_args['--add-package'] = ['vim']
        self.task.process()
        system_prepare.setup_repositories.assert_called_once_with(
            False, ['some_key', 'some_other_key'], None
        )
        system_prepare.install_packages.assert_called_once_with(
            manager.__enter__.return_value, ['vim']
        )

    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.xml_state.XMLState.get_repositories_signing_keys')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_update_delete_package(
        self, mock_SystemPrepare, mock_keys, mock_log
    ):
        manager = MagicMock()
        system_prepare = Mock()
        system_prepare.setup_repositories = Mock(
            return_value=manager
        )
        mock_SystemPrepare.return_value.__enter__.return_value = system_prepare
        mock_keys.return_value = ['some_key', 'some_other_key']
        self._init_command_args()
        self.task.command_args['--delete-package'] = ['vim']
        self.task.process()
        system_prepare.setup_repositories.assert_called_once_with(
            False, ['some_key', 'some_other_key'], None
        )
        system_prepare.delete_packages.assert_called_once_with(
            manager.__enter__.return_value, ['vim']
        )

    @patch('kiwi.xml_state.XMLState.set_container_config_tag')
    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_prepare_stage_set_container_tag(
        self, mock_SystemPrepare, mock_log, mock_set_container_tag
    ):
        self._init_command_args()
        self.task.command_args['--set-container-tag'] = 'new_tag'
        self.task.process()
        mock_set_container_tag.assert_called_once_with(
            'new_tag'
        )

    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_invalid_type_attribute(
        self, mock_SystemPrepare, mock_log
    ):
        self._init_command_args()
        self.task.command_args['--set-type-attr'] = [
            'bogus=value'
        ]
        with self._caplog.at_level(logging.ERROR):
            self.task.process()
        assert 'Failed to set type attribute' in self._caplog.text

    @patch('kiwi.xml_state.XMLState.get_repositories_signing_keys')
    @patch('kiwi.xml_state.XMLState.get_preferences_sections')
    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_release_version_no_overwrite(
        self, mock_SystemPrepare, mock_logger,
        mock_get_preferences_sections, mock_get_repositories_signing_keys
    ):
        preferences = MagicMock()
        preferences.get_release_version.return_value = None
        mock_get_preferences_sections.return_value = [preferences]
        self._init_command_args()
        self.task.command_args['--set-release-version'] = '42'
        self.task.process()
        preferences.add_release_version.assert_called_once_with('42')

    @patch('kiwi.xml_state.XMLState.add_container_config_label')
    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_add_container_label(
        self, mock_SystemPrepare, mock_log, mock_add_container_label
    ):
        self._init_command_args()
        self.task.command_args['--add-container-label'] = [
            'newLabel=value', 'anotherLabel=my=crazy value'
        ]
        self.task.process()
        mock_add_container_label.assert_has_calls([
            call('newLabel', 'value'),
            call('anotherLabel', 'my=crazy value')
        ])

    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_add_container_label_invalid_format(
        self, mock_SystemPrepare, mock_logger
    ):
        self._init_command_args()
        self.task.command_args['--add-container-label'] = ['newLabel:value']
        with self._caplog.at_level(logging.WARNING):
            self.task.process()

    @patch('kiwi.xml_state.XMLState.set_derived_from_image_uri')
    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_prepare_stage_set_derived_from_uri(
        self, mock_SystemPrepare, mock_log, mock_set_derived_from_uri
    ):
        self._init_command_args()
        self.task.command_args['--set-container-derived-from'] = 'file:///new'
        self.task.process()
        mock_set_derived_from_uri.assert_called_once_with(
            'file:///new'
        )

    @patch('kiwi.xml_state.XMLState.set_repository')
    @patch('kiwi.logger.Logger.set_logfile')
    @patch('os.path.isfile')
    @patch('os.unlink')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_prepare_stage_set_repo(
        self, mock_SystemPrepare, mock_os_unlink, mock_os_path_is_file,
        mock_log, mock_set_repo
    ):
        mock_os_path_is_file.return_value = False
        self._init_command_args()
        self.task.command_args['--set-repo'] = 'http://example.com,yast2,alias'
        self.task.process()
        mock_set_repo.assert_called_once_with(
            'http://example.com', 'yast2', 'alias',
            None, None, None, [], None, None, None, None
        )
        self.task.command_args['--set-repo-credentials'] = 'user:pass'
        mock_set_repo.reset_mock()
        self.task.process()
        mock_set_repo.assert_called_once_with(
            'http://user:pass@example.com', 'yast2', 'alias',
            None, None, None, [], None, None, None, None
        )
        self.task.command_args['--set-repo-credentials'] = '../data/credentials'
        mock_os_path_is_file.return_value = True
        mock_set_repo.reset_mock()
        self.task.process()
        mock_set_repo.assert_called_once_with(
            'http://user:pass@example.com', 'yast2', 'alias',
            None, None, None, [], None, None, None, None
        )
        mock_os_unlink.assert_called_once_with('../data/credentials')

    @patch('kiwi.xml_state.XMLState.add_repository')
    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_build_prepare_stage_add_repo(
        self, mock_SystemPrepare, mock_log, mock_add_repo
    ):
        self._init_command_args()
        self.task.command_args['--add-repo'] = [
            'http://example1.com,yast2,alias,99,false,true',
            'http://example2.com,yast2,alias,99,false,true',
            'http://example3.com,yast2,alias,99,false,true',
            'http://example4.com,,alias,99,false,true'
        ]
        self.task.process()
        assert mock_add_repo.call_args_list == [
            call(
                'http://example1.com', 'yast2', 'alias', '99',
                False, True, [], None, None, None, None
            ),
            call(
                'http://example2.com', 'yast2', 'alias', '99',
                False, True, [], None, None, None, None
            ),
            call(
                'http://example3.com', 'yast2', 'alias', '99',
                False, True, [], None, None, None, None
            ),
            call(
                'http://example4.com', None, 'alias', '99',
                False, True, [], None, None, None, None
            )
        ]
        self.task.command_args['--add-repo-credentials'] = [
            'user1:pass1',
            'user2:pass2'
        ]
        mock_add_repo.reset_mock()
        self.task.process()
        assert mock_add_repo.call_args_list == [
            call(
                'http://user1:pass1@example1.com', 'yast2', 'alias', '99',
                False, True, [], None, None, None, None
            ),
            call(
                'http://user2:pass2@example2.com', 'yast2', 'alias', '99',
                False, True, [], None, None, None, None
            ),
            call(
                'http://example3.com', 'yast2', 'alias', '99',
                False, True, [], None, None, None, None
            ),
            call(
                'http://example4.com', None, 'alias', '99',
                False, True, [], None, None, None, None
            )
        ]

    def test_process_system_build_help(self):
        self._init_command_args()
        self.task.command_args['help'] = True
        self.task.command_args['build'] = True
        self.task.process()
        self.task.manual.show.assert_called_once_with(
            'kiwi::system::build'
        )

    @patch('kiwi.xml_state.XMLState.delete_repository_sections')
    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_prepare_ignore_repos(
        self, mock_SystemPrepare, mock_log, mock_delete_repos
    ):
        self._init_command_args()
        self.task.command_args['--ignore-repos'] = True
        self.task.process()
        mock_delete_repos.assert_called_once_with()

    @patch('kiwi.xml_state.XMLState.delete_repository_sections_used_for_build')
    @patch('kiwi.logger.Logger.set_logfile')
    @patch('kiwi.tasks.system_build.SystemPrepare')
    def test_process_system_prepare_ignore_repos_used_for_build(
        self, mock_SystemPrepare, mock_log, mock_delete_repos
    ):
        self._init_command_args()
        self.task.command_args['--ignore-repos-used-for-build'] = True
        self.task.process()
        mock_delete_repos.assert_called_once_with()



from pytest import raises
from unittest.mock import Mock, patch

from kiwi.oci_tools.base import OCIBase


class TestOCIBase:
    def setup(self):
        self.oci = OCIBase()

    def setup_method(self, cls):
        self.setup()

    def test_init_container(self):
        with raises(NotImplementedError):
            self.oci.init_container()

    def test_import_container_image(self):
        with raises(NotImplementedError):
            self.oci.import_container_image('oci-archive:image.xz')

    def test_export_container_image(self):
        with raises(NotImplementedError):
            self.oci.export_container_image(
                'image.xz', 'docker-archive', 'myimage:tag'
            )

    def test_unpack(self):
        with raises(NotImplementedError):
            self.oci.unpack()

    def test_sync_rootfs(self):
        with raises(NotImplementedError):
            self.oci.sync_rootfs('root_dir')

    def test_import_rootfs(self):
        with raises(NotImplementedError):
            self.oci.import_rootfs('root_dir')

    def test_repack(self):
        with raises(NotImplementedError):
            self.oci.repack({})

    def test_set_config(self):
        with raises(NotImplementedError):
            self.oci.set_config({})

    def test_post_process(self):
        with raises(NotImplementedError):
            self.oci.post_process()

    @patch('kiwi.oci_tools.base.Path.which')
    @patch('kiwi.oci_tools.base.CommandCapabilities.check_version')
    def test_skopeo_provides_tmpdir_option(
        self, mock_Path_which, mock_CommandCapabilities_check_version
    ):
        mock_Path_which.return_value = 'skopeo'
        mock_CommandCapabilities_check_version.return_value = (0, 2, 0)
        assert self.oci._skopeo_provides_tmpdir_option() is True
        mock_Path_which.return_value = None
        assert self.oci._skopeo_provides_tmpdir_option() is False

    @patch('kiwi.oci_tools.base.datetime')
    def test_creation_date(self, mock_datetime):
        strftime = Mock()
        strftime.strftime = Mock(return_value='current_date')
        mock_datetime.utcnow = Mock(
            return_value=strftime
        )

        with patch.dict('os.environ', {'SOURCE_DATE_EPOCH': ''}):
            oci = OCIBase()
            assert oci.creation_date == 'current_date'

    def test_creation_date_from_source_date_epoch(self):
        with patch.dict('os.environ', {'SOURCE_DATE_EPOCH': '946729830'}):
            oci = OCIBase()
            assert oci.creation_date == '2000-01-01T12:30:30+00:00'



# Copyright (c) 2022 Marcus Schäfer.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import logging
from typing import Dict

# project
from kiwi.bootloader.install.base import BootLoaderInstallBase

log = logging.getLogger('kiwi')


class BootLoaderInstallSystemdBoot(BootLoaderInstallBase):
    """
    **systemd-boot bootloader installation**
    """
    def post_init(self, custom_args: Dict):
        """
        systemd-boot post initialization method

        :param dict custom_args:
            Contains custom systemd-boot bootloader arguments
        """
        self.custom_args = custom_args

    def install_required(self):
        """
        Check if systemd-boot needs to install boot code

        systemd-boot supports EFI boot only and does not need to
        install boot code since it's expected that the firmware
        can read from the EFI partition

        :return: True or False

        :rtype: bool
        """
        return False

    def secure_boot_install(self):
        """
        Run shim installation for secure boot setup

        For systemd-boot this is currently skipped since details for
        secure boot are not yet clear
        """
        pass



import io
from pytest import raises
from unittest.mock import (
    Mock, patch, call, MagicMock
)
from kiwi.bootloader.config.zipl import BootLoaderZipl
from kiwi.bootloader.config.bootloader_spec_base import BootLoaderSpecBase
from kiwi.command import CommandT

from kiwi.exceptions import (
    KiwiTemplateError,
    KiwiBootLoaderTargetError,
    KiwiDiskGeometryError
)


class TestBootLoaderZipl:
    @patch('kiwi.bootloader.config.bootloader_spec_base.FirmWare')
    def setup(self, mock_FirmWare):
        self.firmware = Mock()
        mock_FirmWare.return_value = self.firmware
        self.state = Mock()
        self.state.get_host_key_certificates.return_value = [
            {
                'hkd_cert': ['/path/to/host.crt'],
                'hkd_revocation_list': ['/path/to/revocation-list.crl'],
                'hkd_ca_cert': '/path/to/DigiCertCA.crt',
                'hkd_sign_cert': '/path/to/ibm-z-host-key-signing.crt'
            }
        ]
        self.state.get_build_type_bootloader_targettype.return_value = 'CDL'
        self.state.build_type.get_target_blocksize.return_value = 4096
        self.state.xml_data.get_name.return_value = 'image-name'
        self.state.get_image_version.return_value = 'image-version'
        self.state.build_type.get_efifatimagesize.return_value = None
        self.bootloader = BootLoaderZipl(self.state, 'root_dir')
        self.bootloader.cmdline = 'options'
        self.bootloader.custom_args['kernel'] = None
        self.bootloader.custom_args['initrd'] = None
        self.bootloader.custom_args['boot_options'] = {}
        self.bootloader.custom_args['targetbase'] = '/dev/disk'
        self.bootloader.custom_args['target_dir'] = 'target_dir'
        self.bootloader.sys_mount = Mock(
            mountpoint='sys_mount'
        )
        self.bootloader.root_mount = Mock(
            mountpoint='system_root_mount'
        )
        self.bootloader._mount_system = Mock()
        self.bootloader.create_efi_path = Mock()
        self.bootloader.get_boot_path = Mock(
            return_value='bootpath'
        )
        self.bootloader.get_menu_entry_title = Mock(
            return_value='title'
        )

    @patch('kiwi.bootloader.config.bootloader_spec_base.FirmWare')
    def setup_method(self, cls, mock_FirmWare):
        self.setup()

    @patch('os.unlink')
    def test_setup_loader_raises_invalid_target(self, mock_os_unlink):
        with raises(KiwiBootLoaderTargetError):
            self.bootloader.setup_loader('iso')

    @patch('shutil.move')
    @patch('os.unlink')
    @patch('os.path.exists')
    @patch('kiwi.bootloader.config.zipl.BootImageBase.get_boot_names')
    @patch('kiwi.bootloader.config.zipl.Path.wipe')
    @patch('kiwi.bootloader.config.zipl.Path.create')
    @patch('kiwi.bootloader.config.zipl.Command.run')
    @patch('kiwi.bootloader.config.zipl.BootLoaderTemplateZipl')
    @patch('kiwi.bootloader.config.zipl.Temporary.new_file')
    @patch.object(BootLoaderZipl, '_write_config_file')
    @patch.object(BootLoaderZipl, '_get_template_parameters')
    @patch.object(BootLoaderSpecBase, 'get_entry_name')
    def test_setup_loader(
        self, mock_get_entry_name, mock_get_template_parameters,
        mock_write_config_file, mock_Temporary_new_file,
        mock_BootLoaderTemplateZipl, mock_Command_run,
        mock_Path_create, mock_Path_wipe, mock_BootImageBase_get_boot_names,
        mock_os_path_exists, mock_os_unlink, mock_shutil_move
    ):
        temporary = Mock()
        temporary.name = 'system_root_mount/kiwi_zipl.conf_'
        mock_Temporary_new_file.return_value.__enter__.return_value = temporary
        mock_get_entry_name.return_value = \
            'opensuse-leap-5.3.18-59.10-default.conf'

        mock_get_template_parameters.return_value = {
            'targetbase': '/dev/loop'
        }
        kernel_info = Mock()
        kernel_info.kernel_version = 'kernel-version'
        kernel_info.kernel_filename = 'kernel-filename'
        kernel_info.initrd_name = 'initrd-name'
        mock_os_path_exists.return_value = True
        mock_BootImageBase_get_boot_names.return_value = kernel_info
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.bootloader.setup_loader('disk')
        file_handle.write.assert_called_once_with('options\n')
        assert mock_write_config_file.call_args_list == [
            call(
                mock_BootLoaderTemplateZipl.
                return_value.get_loader_template.return_value,
                'system_root_mount/kiwi_zipl.conf_',
                mock_get_template_parameters.return_value
            ),
            call(
                mock_BootLoaderTemplateZipl.
                return_value.get_entry_template.return_value,
                'system_root_mount/boot/loader/entries/'
                'opensuse-leap-5.3.18-59.10-default.conf',
                mock_get_template_parameters.return_value
            )
        ]
        assert self.bootloader._mount_system.called
        assert self.bootloader.sys_mount.umount.called
        assert mock_Command_run.call_args_list == [
            call(
                [
                    'chroot', 'system_root_mount', 'genprotimg',
                    '--offline', '--verbose',
                    '-o', 'bootpath/kernel-filename.cc',
                    '-i', 'bootpath/kernel-filename',
                    '-r', 'bootpath/initrd-name',
                    '-p', temporary.name.replace('system_root_mount', ''),
                    '--cert', '/path/to/DigiCertCA.crt',
                    '--cert', '/path/to/ibm-z-host-key-signing.crt',
                    '-k', '/path/to/host.crt',
                    '--crl', '/path/to/revocation-list.crl'
                ]
            ),
            call(
                [
                    'chroot', 'system_root_mount', 'genprotimg',
                    '--offline', '--verbose',
                    '-o', 'bootpath/kernel-filename.cc',
                    '-i', 'bootpath/kernel-filename',
                    '-r', 'bootpath/initrd-name',
                    '-p', temporary.name.replace('system_root_mount', ''),
                    '--no-verify',
                    '-k', '/path/to/host.crt',
                    '--crl', '/path/to/revocation-list.crl',
                ]
            ),
            call(
                [
                    'chroot', 'system_root_mount', 'pvextract-hdr',
                    '-o', 'secure_execution_header.bin',
                    'bootpath/kernel-filename.cc'
                ]
            ),
            call(
                [
                    'chroot', 'system_root_mount', 'pvattest', 'create',
                    '--no-verify', '--verbose',
                    '-o', 'attestation_request.bin',
                    '--arpk', 'attestation_response.key',
                    '-k', '/path/to/host.crt'
                ]
            ),
            call(
                [
                    'chroot', 'system_root_mount', 'zipl',
                    '--noninteractive',
                    '--config', '/kiwi_zipl.conf_',
                    '--blsdir', '/boot/loader/entries',
                    '--verbose'
                ]
            )
        ]
        assert mock_shutil_move.call_args_list == [
            call(
                'system_root_mount/secure_execution_header.bin',
                'target_dir/image-name.x86_64-image-version.'
                'secure_execution_header.bin'
            ),
            call(
                'system_root_mount/attestation_request.bin',
                'target_dir/image-name.x86_64-image-version.'
                'attestation_request.bin'
            ),
            call(
                'system_root_mount/attestation_response.key',
                'target_dir/image-name.x86_64-image-version.'
                'attestation_response.key'
            )
        ]
        assert mock_os_unlink.call_args_list == [
            call('system_root_mount/bootpath/kernel-filename.cc'),
            call('system_root_mount/bootpath/kernel-filename'),
            call('system_root_mount/bootpath/initrd-name')
        ]

    @patch.object(BootLoaderZipl, '_get_disk_geometry')
    @patch.object(BootLoaderZipl, '_get_partition_start')
    def test_get_template_parameters(
        self, mock_get_partition_start, mock_get_disk_geometry
    ):
        mock_get_partition_start.return_value = '42'
        mock_get_disk_geometry.return_value = '123,53,9'
        self.bootloader.timeout = 0
        self.bootloader.disk_type = 'CDL'
        self.bootloader.disk_blocksize = 4096
        assert self.bootloader._get_template_parameters() == {
            'secure_image_file': 'bootpath/kernel-filename.cc',
            'kernel_file': 'vmlinuz',
            'initrd_file': 'initrd',
            'boot_options': 'options',
            'boot_timeout': 0,
            'bootpath': 'bootpath',
            'targetbase': 'targetbase=/dev/disk',
            'targettype': 'CDL',
            'targetblocksize': '4096',
            'targetoffset': '42',
            'targetgeometry': 'targetgeometry=123,53,9',
            'title': 'title',
            'default_entry': ''
        }

    @patch('kiwi.bootloader.config.zipl.Path.create')
    @patch.object(BootLoaderZipl, '_get_template_parameters')
    def test_write_config_file(
        self, mock_get_template_parameters, mock_Path_create
    ):
        template = Mock()
        template.substitute.return_value = 'data'
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.bootloader._write_config_file(template, 'path/some-file', {})
            mock_Path_create.assert_called_once_with('path')
            mock_open.assert_called_once_with('path/some-file', 'w')
            file_handle.write.assert_called_once_with('data')

    def test_write_config_file_raises(self):
        template = Mock()
        template.substitute.side_effect = Exception
        with raises(KiwiTemplateError):
            self.bootloader._write_config_file(template, 'some-file', {})

    @patch('kiwi.bootloader.config.zipl.Command.run')
    def test_get_disk_geometry(self, mock_Command_run):
        command_return_values = [
            CommandT(
                output='  cylinders ............: 10017\n',
                error='', returncode=0
            ),
            CommandT(
                output='  tracks per cylinder ..: 15\n',
                error='', returncode=0
            ),
            CommandT(
                output='  blocks per track .....: 12\n',
                error='', returncode=0
            )
        ]

        def command_run(arg):
            return command_return_values.pop(0)

        self.firmware.get_partition_table_type.return_value = 'dasd'
        mock_Command_run.side_effect = command_run
        assert self.bootloader._get_disk_geometry() == '10017,15,12'

    @patch('kiwi.bootloader.config.zipl.Command.run')
    def test_get_partition_start_dasd(self, mock_Command_run):
        self.firmware.get_partition_table_type.return_value = 'dasd'
        command_return_values = [
            CommandT(
                output='  blocks per track .....: 12\n',
                error='', returncode=0
            ),
            CommandT(
                output=' /dev/loop01 2 6401 6400 1 Linux native\n',
                error='', returncode=0
            )
        ]

        def command_run(arg):
            return command_return_values.pop(0)

        mock_Command_run.side_effect = command_run
        assert self.bootloader._get_partition_start() == '24'
        assert mock_Command_run.call_args_list == [
            call(
                [
                    'bash', '-c',
                    'fdasd -f -p /dev/disk | grep "blocks per track"'
                ]
            ),
            call(
                [
                    'bash', '-c',
                    'fdasd -f -s -p /dev/disk | grep "^ " |'
                    ' head -n 1 | tr -s " "'
                ]
            )
        ]

    @patch('kiwi.bootloader.config.zipl.Command.run')
    def test_get_partition_start_not_dasd(self, mock_Command_run):
        self.firmware.get_partition_table_type.return_value = 'msdos'
        command = Mock()
        command.output = '  2048'
        mock_Command_run.return_value = command
        assert self.bootloader._get_partition_start() == '2048'
        mock_Command_run.assert_called_once_with(
            [
                'bash', '-c',
                'sfdisk --dump /dev/disk | grep "1 :" |'
                ' cut -f1 -d, | cut -f2 -d='
            ]
        )

    @patch('kiwi.bootloader.config.zipl.Command.run')
    @patch.object(BootLoaderZipl, '_get_dasd_disk_geometry_element')
    def test_get_partition_start_raises(
        self, mock_get_dasd_disk_geometry_element, mock_Command_run
    ):
        self.firmware.get_partition_table_type.return_value = 'dasd'
        mock_Command_run.return_value = CommandT(
            output='bogus data', error='', returncode=0
        )
        with raises(KiwiDiskGeometryError):
            self.bootloader._get_partition_start()

    @patch('kiwi.bootloader.config.zipl.Command.run')
    def test_get_dasd_disk_geometry_element_raises(
        self, mock_Command_run
    ):
        self.bootloader.target_table_type = 'dasd'
        mock_Command_run.return_value = CommandT(
            output='bogus data', error='', returncode=0
        )
        with raises(KiwiDiskGeometryError):
            self.bootloader._get_dasd_disk_geometry_element(
                '/dev/disk', 'tracks per cylinder'
            )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
"""
usage: kiwi-ng image info -h | --help
       kiwi-ng image info --description=<directory>
           [--resolve-package-list]
           [--list-profiles]
           [--print-kiwi-env]
           [--ignore-repos]
           [--add-repo=<source,type,alias,priority>...]
           [--print-xml|--print-yaml|--print-toml]
       kiwi-ng image info help

commands:
    info
        provide information about the specified image description

options:
    --add-repo=<source,type,alias,priority>
        add repository with given source, type, alias and priority
    --description=<directory>
        the description must be a directory containing a kiwi XML
        description and optional metadata files
    --ignore-repos
        ignore all repos from the XML configuration
    --resolve-package-list
        solve package dependencies and return a list of all
        packages including their attributes e.g size,
        shasum, etc...
    --list-profiles
        list profiles available for the selected/default type
    --print-kiwi-env
        print kiwi profile environment variables
    --print-xml|--print-yaml|--print-toml
        print image description in specified format
"""
import os

# project
from kiwi.tasks.base import CliTask
from kiwi.help import Help
from kiwi.utils.output import DataOutput
from kiwi.solver.sat import Sat
from kiwi.solver.repository import SolverRepository
from kiwi.solver.repository.base import SolverRepositoryBase
from kiwi.system.uri import Uri
from kiwi.system.profile import Profile
from kiwi.defaults import Defaults


class ImageInfoTask(CliTask):
    """
    Implements retrieval of in depth information for an image description

    Attributes

    * :attr:`manual`
        Instance of Help
    """
    def process(self):
        """
        Walks through the given info options and provide the requested data
        """
        self.manual = Help()
        if self.command_args.get('help') is True:
            return self.manual.show('kiwi::image::info')

        self.load_xml_description(
            self.command_args['--description'], self.global_args['--kiwi-file']
        )

        if self.command_args['--ignore-repos']:
            self.xml_state.delete_repository_sections()

        if self.command_args['--add-repo']:
            for add_repo in self.command_args['--add-repo']:
                (repo_source, repo_type, repo_alias, repo_prio) = \
                    self.quadruple_token(add_repo)
                self.xml_state.add_repository(
                    repo_source, repo_type, repo_alias, repo_prio
                )

        self.runtime_checker.check_repositories_configured()

        result = {
            'image': self.xml_state.xml_data.get_name()
        }

        if self.command_args['--print-kiwi-env']:
            profile = Profile(self.xml_state)
            defaults = Defaults()
            defaults.to_profile(profile)
            result['kiwi_env'] = profile.get_settings()

        if self.command_args['--list-profiles']:
            result['profile_names'] = []
            for profiles_section in self.xml_state.xml_data.get_profiles():
                for profile in profiles_section.get_profile():
                    result['profile_names'].append(profile.get_name())

        if self.command_args['--resolve-package-list']:
            solver = self._setup_solver()
            boostrap_package_list = self.xml_state.get_bootstrap_packages()
            package_list = boostrap_package_list + \
                self.xml_state.get_system_packages()
            bootstrap_collection_type = \
                self.xml_state.get_bootstrap_collection_type()
            system_collection_type = \
                self.xml_state.get_system_collection_type()
            bootstrap_packages = solver.solve(
                boostrap_package_list, False,
                True if bootstrap_collection_type == 'onlyRequired' else False
            )
            solved_packages = solver.solve(
                self.xml_state.get_system_packages(), False,
                True if system_collection_type == 'onlyRequired' else False
            )
            solved_packages.update(bootstrap_packages)
            package_info = {}
            for package, metadata in sorted(list(solved_packages.items())):
                if package in package_list:
                    status = 'listed_in_kiwi_description'
                else:
                    status = 'added_by_dependency_solver'
                package_info[package] = {
                    'source': metadata.uri,
                    'installsize_bytes': int(metadata.installsize_bytes),
                    'arch': metadata.arch,
                    'version': metadata.version,
                    'status': status
                }
            result['resolved-packages'] = package_info

        if self.global_args['--color-output']:
            DataOutput(result, style='color').display()
        else:
            DataOutput(result).display()

        if self.command_args['--print-xml']:
            DataOutput.display_file(
                self.description.markup.get_xml_description(),
                'Description(XML):'
            )
        elif self.command_args['--print-yaml']:
            DataOutput.display_file(
                self.description.markup.get_yaml_description(),
                'Description(YAML):'
            )
        elif self.command_args['--print-toml']:
            DataOutput.display_file(
                self.description.markup.get_toml_description(),
                'Description(TOML):'
            )

    def _setup_solver(self):
        solver = Sat()
        for xml_repo in self.xml_state.get_repository_sections_used_for_build():
            repo_source = xml_repo.get_source().get_path()
            repo_sourcetype = xml_repo.get_sourcetype() or ''
            repo_user = xml_repo.get_username()
            repo_secret = xml_repo.get_password()
            repo_type = xml_repo.get_type()
            repo_dist = xml_repo.get_distribution()
            repo_components = xml_repo.get_components()
            if not repo_type:
                repo_type = SolverRepositoryBase(
                    Uri(uri=repo_source, source_type=repo_sourcetype),
                    repo_user, repo_secret
                ).get_repo_type()
            if repo_type == 'apt-deb':
                # Debian based repos can be setup for a specific
                # distribution including a list of individual components.
                # For each component of the selected distribution extra
                # repository metadata exists. In such a case we iterate
                # over the configured dist components and add them as
                # repository each.
                dist_type = solver.set_dist_type('deb')
                if repo_components and repo_dist:
                    for component in repo_components.split():
                        repo_source_for_component = os.sep.join(
                            [
                                repo_source.rstrip(os.sep), 'dists', repo_dist,
                                component, f'binary-{dist_type.get("arch")}'
                            ]
                        )
                        solver.add_repository(
                            SolverRepository.new(
                                Uri(
                                    repo_source_for_component,
                                    repo_type, repo_sourcetype
                                ),
                                repo_user, repo_secret
                            )
                        )
                    continue
            solver.add_repository(
                SolverRepository.new(
                    Uri(repo_source, repo_type, repo_sourcetype),
                    repo_user, repo_secret
                )
            )
        return solver



from unittest.mock import (
    patch, Mock
)
from pytest import raises

from kiwi.package_manager import PackageManager

from kiwi.exceptions import KiwiPackageManagerSetupError


class TestPackageManager:
    def test_package_manager_not_implemented(self):
        with raises(KiwiPackageManagerSetupError):
            PackageManager.new('repository', 'ms-manager')

    @patch('kiwi.package_manager.zypper.PackageManagerZypper')
    def test_manager_zypper(self, mock_manager):
        repository = Mock()
        PackageManager.new(repository, 'zypper')
        mock_manager.assert_called_once_with(repository, [], '')

    @patch('kiwi.package_manager.dnf.PackageManagerDnf')
    def test_manager_dnf(self, mock_manager):
        repository = Mock()
        PackageManager.new(repository, 'dnf')
        mock_manager.assert_called_once_with(repository, [], '')

    @patch('kiwi.package_manager.microdnf.PackageManagerMicroDnf')
    def test_manager_microdnf(self, mock_manager):
        repository = Mock()
        PackageManager.new(repository, 'microdnf')
        mock_manager.assert_called_once_with(repository, [], '')

    @patch('kiwi.package_manager.apt.PackageManagerApt')
    def test_manager_apt(self, mock_manager):
        repository = Mock()
        PackageManager.new(repository, 'apt')
        mock_manager.assert_called_once_with(repository, [], '')

    @patch('kiwi.package_manager.pacman.PackageManagerPacman')
    def test_manager_pacman(self, mock_manager):
        repository = Mock()
        PackageManager.new(repository, 'pacman')
        mock_manager.assert_called_once_with(repository, [], '')



from unittest.mock import Mock
from pytest import raises

from kiwi.partitioner.base import PartitionerBase


class TestPartitionerBase:
    def setup(self):
        disk_provider = Mock()
        disk_provider.get_device = Mock(
            return_value='/dev/loop0'
        )
        self.partitioner = PartitionerBase(disk_provider)

    def setup_method(self, cls):
        self.setup()

    def test_get_id(self):
        assert self.partitioner.get_id() == 0

    def test_set_uuid(self):
        with raises(NotImplementedError):
            self.partitioner.set_uuid(100, 'ID')

    def test_create(self):
        with raises(NotImplementedError):
            self.partitioner.create('name', 100, 'type', ['flag'])

    def test_set_flag(self):
        with raises(NotImplementedError):
            self.partitioner.set_flag(1, 'flag')

    def test_set_hybrid_mbr(self):
        with raises(NotImplementedError):
            self.partitioner.set_hybrid_mbr()

    def test_set_mbr(self):
        with raises(NotImplementedError):
            self.partitioner.set_mbr()

    def test_resize_table(self):
        with raises(NotImplementedError):
            self.partitioner.resize_table()

    def test_set_start_sector(self):
        assert self.partitioner.set_start_sector(4096) is None



from pytest import raises

from kiwi.utils.os_release import OsRelease
from kiwi.exceptions import KiwiOSReleaseImportError


class TestOsRelease(object):
    def setup(self):
        self.os_release = OsRelease('../data')

    def setup_method(self, cls):
        self.setup()

    def test_setup_raises(self):
        with raises(KiwiOSReleaseImportError):
            OsRelease('does-not-exist')

    def test_get(self):
        assert self.os_release.get('ID') == 'opensuse-leap'



# Copyright (c) 2021 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#

import math

from typing import List, Iterator

# initial list of prime numbers which will be extended as much as required
# and reused in next calls

_primes: List[int] = [2, 3, 5, 7, 11, 13, 17]


def _update_primes(n: int) -> None:
    """
    Extend the list of prime numbers if required.
    """
    if n <= _primes[-1]:
        return
    _n = int(math.ceil(n**0.5))
    _update_primes(_n)

    start = _primes[-1] + 1
    for i in range(start, n + 1):
        if all(i % p for p in primes(_n)):
            _primes.append(i)


def primes(number: int) -> Iterator[int]:
    """
    Get prime numbers no greater than given number.

    :param int number: highest possible number to return.

    :rtype: int generator
    """
    _update_primes(number)
    for p in _primes:
        if p <= number:
            yield p
        else:
            break


def factors(number: int, max_factor: int = None) -> Iterator[int]:
    _max = int(math.ceil(number**0.5))
    _max = min(_max, max_factor or _max)
    _number = number
    for i in primes(_max):
        while _number % i == 0 and i <= _max:
            _number = _number // i
            _max = min(_max, int(math.ceil(_number**0.5)))
            yield i
    if _number != 1:
        if not max_factor or _number <= max_factor:
            yield _number



from unittest.mock import (
    patch, Mock
)
from pytest import raises

from kiwi.storage.mapped_device import MappedDevice

from kiwi.exceptions import KiwiMappedDeviceError


class TestMappedDevice:
    @patch('os.path.exists')
    def setup(self, mock_path):
        mock_path.return_value = True
        self.device_provider = Mock()
        self.device_provider.is_loop = Mock()
        self.device = MappedDevice(
            '/dev/foo', self.device_provider
        )

    @patch('os.path.exists')
    def setup_method(self, cls, mock_path):
        self.setup()

    @patch('os.path.exists')
    def test_device_not_existingr(self, mock_path):
        mock_path.return_value = False
        with raises(KiwiMappedDeviceError):
            MappedDevice('/dev/foo', Mock())

    def test_get_device(self):
        assert self.device.get_device() == '/dev/foo'

    def test_is_loop(self):
        assert self.device.is_loop() == self.device_provider.is_loop()



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import logging
from collections import OrderedDict

# project
from kiwi.command import Command
from kiwi.runtime_config import RuntimeConfig
from kiwi.defaults import Defaults
from kiwi.path import Path
from kiwi.xml_state import XMLState
from kiwi.system.result import Result

from kiwi.exceptions import (
    KiwiFormatSetupError,
    KiwiResizeRawDiskError
)

log = logging.getLogger('kiwi')


class DiskFormatBase:
    """
    **Base class to create disk formats from a raw disk image**

    :param object xml_state: Instance of XMLState
    :param string root_dir: root directory path name
    :param string arch: Defaults.get_platform_name
    :param string target_dir: target directory path name
    :param dict custom_args: custom format options dictionary
    """
    def __init__(
            self, xml_state: XMLState, root_dir: str, target_dir: str,
            custom_args: dict = None
    ):
        self.xml_state = xml_state
        self.root_dir = root_dir
        self.arch = Defaults.get_platform_name()
        self.target_dir = target_dir
        self.temp_image_dir: str = ''
        self.image_format: str = ''
        self.diskname = self.get_target_file_path_for_format('raw')
        self.runtime_config = RuntimeConfig()

        self.post_init(custom_args or {})

    def __enter__(self):
        return self

    def post_init(self, custom_args: dict) -> None:
        """
        Post initialization method

        Implementation in specialized disk format class if required

        :param dict custom_args: unused
        """
        pass

    def has_raw_disk(self) -> bool:
        """
        Check if the base raw disk image exists

        :return: True or False

        :rtype: bool
        """
        return os.path.exists(self.diskname)

    def resize_raw_disk(self, size_bytes, append=False):
        """
        Resize raw disk image to specified size. If the request
        would actually shrink the disk an exception is raised.
        If the disk got changed the method returns True, if
        the new size is the same as the current size nothing
        gets resized and the method returns False

        :param int size: size in bytes

        :return: True or False

        :rtype: bool
        """
        if not append:
            current_byte_size = os.path.getsize(self.diskname)
            size_bytes = int(size_bytes)
            if size_bytes < current_byte_size:
                raise KiwiResizeRawDiskError(
                    'shrinking {0} disk to {1} bytes corrupts the image'.format(
                        self.diskname, size_bytes
                    )
                )
            elif size_bytes == current_byte_size:
                return False
        Command.run(
            [
                'qemu-img', 'resize', self.diskname,
                '+{0}'.format(size_bytes) if append else format(size_bytes)
            ]
        )
        return True

    def create_image_format(self) -> None:
        """
        Create disk format

        Implementation in specialized disk format class required
        """
        raise NotImplementedError

    @staticmethod
    def get_qemu_option_list(custom_args: dict) -> list:
        """
        Create list of qemu options from custom_args dict

        :param dict custom_args: arguments

        :return: qemu option list

        :rtype: list
        """
        options = []
        if custom_args:
            ordered_args = OrderedDict(sorted(custom_args.items()))
            for key, value in list(ordered_args.items()):
                if key == 'adapter_type=pvscsi':
                    # Building for the pvscsi ddb.adapterType only
                    # affects the guest configuration (vmx). For the
                    # vmdk disk format this is the same as lsilogic
                    key = 'adapter_type=lsilogic'
                options.append('-o')
                if value:
                    options.append('{0}={1}'.format(key, value))
                else:
                    options.append(key)
        return options

    def get_target_file_path_for_format(self, format_name: str) -> str:
        """
        Create target file path name for specified format

        :param string format_name: disk format name

        :return: file path name

        :rtype: str
        """
        if format_name != 'raw':
            if format_name not in Defaults.get_disk_format_types():
                raise KiwiFormatSetupError(
                    f'unsupported disk format {format_name}'
                )
        return ''.join(
            [
                self.target_dir, '/',
                self.xml_state.xml_data.get_name(),
                '.' + self.arch,
                '-' + self.xml_state.get_image_version(),
                '.' + format_name
            ]
        )

    def store_to_result(self, result: Result):
        """
        Store result file of the format conversion into the
        provided result instance.

        By default only the converted image file will be stored
        as compressed file. Subformats which creates additional
        metadata files or want to use other result flags needs
        to overwrite this method

        :param object result: Instance of Result
        """
        compression = self.runtime_config.get_bundle_compression(default=True)
        if self.xml_state.get_luks_credentials() is not None:
            compression = False
        result.add(
            key='disk_format_image',
            filename=self.get_target_file_path_for_format(
                self.image_format
            ),
            use_for_bundle=True,
            compress=compression,
            shasum=True
        )

    def __exit__(self, exc_type, exc_value, traceback):
        if self.temp_image_dir and os.path.exists(self.temp_image_dir):
            Path.wipe(self.temp_image_dir)



from unittest.mock import patch
from pytest import raises

from kiwi.help import Help

from kiwi.exceptions import KiwiHelpNoCommandGiven


class TestHelp:
    def setup(self):
        self.help = Help()

    def setup_method(self, cls):
        self.setup()

    def test_show(self):
        with raises(KiwiHelpNoCommandGiven):
            self.help.show(None)

    @patch('subprocess.call')
    def test_show_command(self, mock_process):
        self.help.show('foo')
        mock_process.assert_called_once_with('man 8 foo', shell=True)



from unittest.mock import (
    patch, Mock
)
from pytest import raises

from kiwi.exceptions import KiwiBootLoaderConfigSetupError
from kiwi.bootloader.config import create_boot_loader_config


class TestBootLoaderConfig:
    def test_bootloader_config_not_implemented(self):
        with raises(KiwiBootLoaderConfigSetupError):
            create_boot_loader_config(
                name='foo', xml_state=Mock(), root_dir='root_dir'
            )

    @patch('kiwi.bootloader.config.grub2.BootLoaderConfigGrub2')
    def test_bootloader_config_grub2(self, mock_grub2):
        xml_state = Mock()
        for name in ("grub2", "grub2_s390x_emu"):
            create_boot_loader_config(
                name=name, xml_state=xml_state, root_dir='root_dir'
            )
            mock_grub2.assert_called_once_with(
                xml_state, 'root_dir', None, None
            )
            mock_grub2.reset_mock()

    @patch('kiwi.bootloader.config.systemd_boot.BootLoaderSystemdBoot')
    def test_bootloader_config_systemd_boot(self, mock_systemd_boot):
        xml_state = Mock()
        create_boot_loader_config(
            name='systemd_boot', xml_state=xml_state,
            root_dir='root_dir', boot_dir='boot_dir'
        )
        mock_systemd_boot.assert_called_once_with(
            xml_state, 'root_dir', 'boot_dir', None
        )

    @patch('kiwi.bootloader.config.zipl.BootLoaderZipl')
    @patch('kiwi.bootloader.config.bootloader_spec_base.FirmWare')
    def test_bootloader_config_zipl(self, mock_FirmWare, mock_zipl):
        xml_state = Mock()
        create_boot_loader_config(
            name='zipl', xml_state=xml_state,
            root_dir='root_dir', boot_dir='boot_dir'
        )
        mock_zipl.assert_called_once_with(
            xml_state, 'root_dir', 'boot_dir', None
        )

    @patch('kiwi.bootloader.config.custom.BootLoaderConfigCustom')
    def test_bootloader_config_custom(self, mock_custom):
        xml_state = Mock()
        create_boot_loader_config(
            name='custom', xml_state=xml_state, root_dir='root_dir'
        )
        mock_custom.assert_called_once_with(
            xml_state, 'root_dir', None, None
        )



# Copyright (c) 2024 SUSE Software Solutions Germany GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import csv
import os
from io import TextIOWrapper
from typing import Iterable

# project
from kiwi.exceptions import KiwiEnvImportError


class ToEnv:
    """
    **Read env file and merge with os.environ**
    """
    def __init__(self, root_dir: str, envfile: str):
        self.data = {}
        env_file = f'{root_dir}/{envfile}'
        if os.path.isfile(env_file):
            try:
                with open(env_file) as osdata:
                    reader = csv.reader(ToEnv._rip(osdata), delimiter='=')
                    self.data = dict(reader)
            except Exception as issue:
                raise KiwiEnvImportError(
                    f'Import of {envfile} failed with {issue}'
                )
            if self.data:
                os.environ.update(self.data)

    @staticmethod
    def _is_comment(line: str) -> bool:
        return line.startswith('#')

    @staticmethod
    def _is_whitespace(line: str) -> bool:
        return line.isspace()

    @staticmethod
    def _rip(csvfile: TextIOWrapper) -> Iterable[str]:
        for row in csvfile:
            if not ToEnv._is_comment(row) \
               and not ToEnv._is_whitespace(row):
                yield row



import logging
from typing import NoReturn
from pytest import (
    fixture, raises
)
from unittest.mock import (
    patch, call, Mock
)
from kiwi.exceptions import KiwiCommandError, KiwiUmountBusyError
from kiwi.mount_manager import MountManager


class TestMountManager:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    @patch('kiwi.mount_manager.Path.create')
    def setup(self, mock_path_create):
        self.mount_manager = MountManager(
            '/dev/some-device', '/some/mountpoint'
        )
        mock_path_create.assert_called_once_with('/some/mountpoint')

    @patch('kiwi.mount_manager.Path.create')
    def setup_method(self, cls, mock_path_create):
        self.setup()

    def test_get_attributes(self):
        assert self.mount_manager.get_attributes() == {}

    @patch('kiwi.mount_manager.Temporary')
    def test_setup_empty_mountpoint(self, mock_Temporary):
        mock_Temporary.return_value.new_dir.return_value.name = 'tmpdir'
        mount_manager = MountManager('/dev/some-device')
        assert mount_manager.mountpoint == 'tmpdir'

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    def test_bind_mount(self, mock_mounted, mock_command):
        mock_mounted.return_value = False
        self.mount_manager.bind_mount()
        mock_command.assert_called_once_with(
            ['mount', '-n', '--bind', '/dev/some-device', '/some/mountpoint']
        )

    @patch('kiwi.mount_manager.Path.create')
    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    def test_overlay_mount(self, mock_mounted, mock_command, mock_path_create):
        mock_mounted.return_value = False
        self.mount_manager.overlay_mount('lower_path')
        assert mock_path_create.call_args_list == [
            call('/some/mountpoint_cow'), call('/some/mountpoint_work')
        ]
        mock_command.assert_called_once_with(
            [
                'mount', '-t', 'overlay', 'overlay', '/some/mountpoint',
                '-o', 'lowerdir=lower_path,'
                'upperdir=/some/mountpoint_cow,workdir=/some/mountpoint_work'
            ]
        )

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    def test_tmpfs_mount(self, mock_mounted, mock_command):
        mock_mounted.return_value = False
        self.mount_manager.tmpfs_mount()
        mock_command.assert_called_once_with(
            ['mount', '-t', 'tmpfs', 'tmpfs', '/some/mountpoint']
        )

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    def test_mount(self, mock_mounted, mock_command):
        mock_mounted.return_value = False
        self.mount_manager.mount(['options'])
        mock_command.assert_called_once_with(
            ['mount', '-o', 'options', '/dev/some-device', '/some/mountpoint']
        )

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    def test_context_manager(self, mock_mounted, mock_command):
        mock_mounted.return_value = True
        with self.mount_manager:
            pass
        mock_command.assert_called_once_with(
            ['umount', '/some/mountpoint']
        )

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    def test_umount_lazy(self, mock_mounted, mock_command):
        mock_mounted.return_value = True
        self.mount_manager.umount_lazy()
        mock_command.assert_called_once_with(
            ['umount', '-l', '/some/mountpoint']
        )

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    @patch('time.sleep')
    def test_umount_with_errors_but_lazy(
        self, mock_sleep, mock_mounted, mock_command
    ):
        command_errors = [
            True, False, False, False, False, False
        ]

        def _cmd_err(args) -> NoReturn:
            if not command_errors.pop():
                raise KiwiCommandError('error')

        mock_command.side_effect = _cmd_err
        mock_mounted.return_value = True
        with self._caplog.at_level(logging.WARNING):
            assert self.mount_manager.umount() is True
        assert mock_command.call_args_list == [
            call(['umount', '/some/mountpoint']),  # 1
            call(['umount', '/some/mountpoint']),  # 2
            call(['umount', '/some/mountpoint']),  # 3
            call(['umount', '/some/mountpoint']),  # 4
            call(['umount', '/some/mountpoint']),  # 5
            call(['umount', '--lazy', '/some/mountpoint']),  # lazy
        ]

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    @patch('time.sleep')
    def test_umount_with_errors(
        self, mock_sleep, mock_mounted, mock_command
    ):
        def _cmd_err(args) -> NoReturn:
            raise KiwiCommandError('error')

        mock_command.side_effect = _cmd_err
        mock_mounted.return_value = True
        with self._caplog.at_level(logging.WARNING):
            assert self.mount_manager.umount(raise_on_busy=False) is False
        assert mock_command.call_args_list == [
            call(['umount', '/some/mountpoint']),  # 1
            call(['umount', '/some/mountpoint']),  # 2
            call(['umount', '/some/mountpoint']),  # 3
            call(['umount', '/some/mountpoint']),  # 4
            call(['umount', '/some/mountpoint']),  # 5
            call(['umount', '--lazy', '/some/mountpoint']),  # lazy
        ]

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    @patch('time.sleep')
    @patch('kiwi.mount_manager.Path.which')
    def test_umount_with_errors_raises_no_lsof_present(
        self, mock_Path_which, mock_sleep, mock_mounted, mock_command
    ):
        def command_call(args):
            if 'umount' in args:
                raise KiwiCommandError('error')

        mock_Path_which.return_value = None
        mock_command.side_effect = command_call
        mock_mounted.return_value = True
        with raises(KiwiUmountBusyError):
            self.mount_manager.umount()

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    @patch('time.sleep')
    @patch('kiwi.mount_manager.Path.which')
    def test_umount_with_errors_raises_lsof_present(
        self, mock_Path_which, mock_sleep, mock_mounted, mock_command
    ):
        def command_call(args, raise_on_error=None):
            if 'umount' in args:
                raise KiwiCommandError('error')
            else:
                call_return = Mock()
                call_return.output = 'HEADLINE\ndata'
                return call_return

        mock_Path_which.return_value = 'lsof'
        mock_command.side_effect = command_call
        mock_mounted.return_value = True
        with raises(KiwiUmountBusyError) as issue:
            self.mount_manager.umount()
        assert 'HEADLINE' in issue.value.message

    @patch('kiwi.mount_manager.Command.run')
    @patch('kiwi.mount_manager.MountManager.is_mounted')
    def test_umount_success(self, mock_mounted, mock_command):
        mock_mounted.return_value = True
        assert self.mount_manager.umount() is True
        mock_command.assert_called_once_with(
            ['umount', '/some/mountpoint']
        )

    @patch('kiwi.mount_manager.Command.run')
    def test_is_mounted_true(self, mock_command):
        command = Mock()
        command.returncode = 0
        mock_command.return_value = command
        assert self.mount_manager.is_mounted() is True
        mock_command.assert_called_once_with(
            command=['mountpoint', '-q', '/some/mountpoint'],
            raise_on_error=False
        )

    @patch('kiwi.mount_manager.Command.run')
    def test_is_mounted_false(self, mock_command):
        command = Mock()
        command.returncode = 1
        mock_command.return_value = command
        assert self.mount_manager.is_mounted() is False
        mock_command.assert_called_once_with(
            command=['mountpoint', '-q', '/some/mountpoint'],
            raise_on_error=False
        )



# Copyright (c) 2025 SUSE LLC.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from typing import Dict

# project
from kiwi.archive.tar import ArchiveTar
from kiwi.defaults import Defaults
from kiwi.utils.compress import Compress
from kiwi.command import Command
from kiwi.container.base import ContainerImageBase


class ContainerImageWsl(ContainerImageBase):
    """
    Create new WSL root tar
    WSL(Windows Subsystem Linux >= 2.4.4)

    :param string root_dir: root directory path name
    :param dict custom_args:
        Custom processing arguments defined as hash keys
    """
    def __init__(self, root_dir: str, custom_args: Dict[str, str] = {}):
        self.root_dir = root_dir
        self.wsl_config = custom_args or {}

    def create(
        self, filename: str, base_image: str = '',
        ensure_empty_tmpdirs: bool = False, compress_archive: bool = True
    ) -> str:
        """
        Create WSL root tar

        :param string filename: archive file name
        :param string base_image: not-supported
        :param bool ensure_empty_tmpdirs: not-supported
        :param bool compress_archive: not-supported
        """
        exclude_list = Defaults.\
            get_exclude_list_for_root_data_sync() + Defaults.\
            get_exclude_list_from_custom_exclude_files(self.root_dir)
        exclude_list.append('boot')
        exclude_list.append('dev')
        exclude_list.append('sys')
        exclude_list.append('proc')

        archive_file_name = filename
        archive = ArchiveTar(archive_file_name)
        archive_file_name = archive.create(
            self.root_dir, exclude=exclude_list,
            options=['--numeric-owner', '--absolute-names']
        )
        compressor = Compress(archive_file_name)
        archive_file_name = compressor.gzip()

        Command.run(
            ['mv', archive_file_name, filename]
        )
        return filename



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import sys
from typing import TYPE_CHECKING, Dict, Union, overload

if sys.version_info >= (3, 8):
    from typing import Literal  # pragma: no cover
else:  # pragma: no cover
    from typing_extensions import Literal  # pragma: no cover

from kiwi.exceptions import (
    KiwiBootLoaderConfigSetupError
)

if TYPE_CHECKING:  # pragma: nocover
    from kiwi.bootloader.config.grub2 import BootLoaderConfigGrub2
    from kiwi.bootloader.config.systemd_boot import BootLoaderSystemdBoot
    from kiwi.bootloader.config.custom import BootLoaderConfigCustom
    from kiwi.bootloader.config.zipl import BootLoaderZipl


@overload
def create_boot_loader_config(
    *, name: Literal["grub2"], xml_state: object, root_dir: str,
    boot_dir: str = None, custom_args: Dict = None
) -> "BootLoaderConfigGrub2":
    ...  # pragma: nocover


@overload
def create_boot_loader_config(
    *, name: Literal["grub2_s390x_emu"], xml_state: object, root_dir: str,
    boot_dir: str = None, custom_args: Dict = None
) -> "BootLoaderConfigGrub2":
    ...  # pragma: nocover


@overload
def create_boot_loader_config(
    *, name: Literal["systemd_boot"], xml_state: object, root_dir: str,
    boot_dir: str = None, custom_args: Dict = None
) -> "BootLoaderSystemdBoot":
    ...  # pragma: nocover


@overload
def create_boot_loader_config(
    *, name: Literal["zipl"], xml_state: object, root_dir: str,
    boot_dir: str = None, custom_args: Dict = None
) -> "BootLoaderZipl":
    ...  # pragma: nocover


@overload
def create_boot_loader_config(
    *, name: str, xml_state: object, root_dir: str,
    boot_dir: str = None, custom_args: Dict = None
) -> "Union[BootLoaderConfigGrub2, BootLoaderSystemdBoot, BootLoaderZipl]":
    ...  # pragma: nocover


def create_boot_loader_config(
    *, name: str, xml_state: object, root_dir: str,
    boot_dir: str = None, custom_args: Dict = None
) -> "Union[BootLoaderConfigGrub2, BootLoaderSystemdBoot, BootLoaderZipl, BootLoaderConfigCustom]":

    if name in ("grub2", "grub2_s390x_emu"):
        from kiwi.bootloader.config.grub2 import BootLoaderConfigGrub2
        return BootLoaderConfigGrub2(xml_state, root_dir, boot_dir, custom_args)
    if name == "systemd_boot":
        from kiwi.bootloader.config.systemd_boot import BootLoaderSystemdBoot
        return BootLoaderSystemdBoot(xml_state, root_dir, boot_dir, custom_args)
    if name == "zipl":
        from kiwi.bootloader.config.zipl import BootLoaderZipl
        return BootLoaderZipl(xml_state, root_dir, boot_dir, custom_args)
    if name == "custom":
        from kiwi.bootloader.config.custom import BootLoaderConfigCustom
        return BootLoaderConfigCustom(xml_state, root_dir, boot_dir, custom_args)

    raise KiwiBootLoaderConfigSetupError(
        f'Support for {name} bootloader config not implemented'
    )



import logging
from unittest.mock import patch
from pytest import fixture
import unittest.mock as mock

from kiwi.filesystem.setup import FileSystemSetup


class TestFileSystemSetup:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    @patch('kiwi.filesystem.setup.SystemSize')
    def setup(self, mock_size):
        size = mock.Mock()
        size.accumulate_mbyte_file_sizes = mock.Mock(
            return_value=42
        )
        size.customize = mock.Mock(
            return_value=42
        )
        mock_size.return_value = size
        self.xml_state = mock.Mock()
        self.xml_state.get_build_type_name = mock.Mock(
            return_value='ext4'
        )
        self.xml_state.get_build_type_unpartitioned_bytes = mock.Mock(
            return_value=0
        )
        self.setup = FileSystemSetup(
            self.xml_state, 'root_dir'
        )

    @patch('kiwi.filesystem.setup.SystemSize')
    def setup_method(self, cls, mock_size):
        self.setup()

    def test_init_with_unpartitioned(self):
        self.xml_state.get_build_type_unpartitioned_bytes = mock.Mock(
            return_value=1024
        )
        FileSystemSetup(self.xml_state, 'root_dir')

    def test_setup_with_pxe_type(self):
        self.xml_state.get_build_type_name = mock.Mock(
            return_value='pxe'
        )
        self.xml_state.build_type.get_filesystem = mock.Mock(
            return_value='xfs'
        )
        setup = FileSystemSetup(
            self.xml_state, 'root_dir'
        )
        assert setup.requested_filesystem == 'xfs'

    def test_get_size_mbytes_calculated(self):
        self.setup.configured_size = None
        assert self.setup.get_size_mbytes() == 42

    def test_get_size_mbytes_configured_additive(self):
        self.setup.configured_size.mbytes = 20
        self.setup.configured_size.additive = True
        assert self.setup.get_size_mbytes() == 62

    def test_get_size_mbytes_configured(self):
        self.setup.configured_size.mbytes = 3
        self.setup.configured_size.additive = False
        with self._caplog.at_level(logging.WARNING):
            assert self.setup.get_size_mbytes() == 3
